package gnmiserver

import (
	"fmt"
	"strconv"
	"strings"
	"time"

	"code.fbi.h-da.de/danet/gnmi-target/handler"
	log "github.com/golang/glog"
	"github.com/openconfig/goyang/pkg/yang"
	"github.com/openconfig/ygot/ygot"

	"github.com/openconfig/gnmi/proto/gnmi"
	pb "github.com/openconfig/gnmi/proto/gnmi"
)

// getChildNode gets a node's child with corresponding schema specified by path
// element. If not found and createIfNotExist is set as true, an empty node is
// created and returned.
func getChildNode(node map[string]interface{}, schema *yang.Entry, elem *pb.PathElem, createIfNotExist bool) (interface{}, *yang.Entry) {
	var nextSchema *yang.Entry
	var ok bool

	if nextSchema, ok = schema.Dir[elem.Name]; !ok {
		return nil, nil
	}

	var nextNode interface{}
	if elem.GetKey() == nil {
		if nextNode, ok = node[elem.Name]; !ok {
			if createIfNotExist {
				node[elem.Name] = make(map[string]interface{})
				nextNode = node[elem.Name]
			}
		}
		return nextNode, nextSchema
	}

	nextNode = getKeyedListEntry(node, elem, createIfNotExist)
	return nextNode, nextSchema
}

// getKeyedListEntry finds the keyed list entry in node by the name and key of
// path elem. If entry is not found and createIfNotExist is true, an empty entry
// will be created (the list will be created if necessary).
func getKeyedListEntry(node map[string]interface{}, elem *pb.PathElem, createIfNotExist bool) map[string]interface{} {
	curNode, ok := node[elem.Name]
	if !ok {
		if !createIfNotExist {
			return nil
		}

		// Create a keyed list as node child and initialize an entry.
		m := make(map[string]interface{})
		for k, v := range elem.Key {
			m[k] = v
			if vAsNum, err := strconv.ParseFloat(v, 64); err == nil {
				m[k] = vAsNum
			}
		}
		node[elem.Name] = []interface{}{m}
		return m
	}

	// Search entry in keyed list.
	keyedList, ok := curNode.([]interface{})
	if !ok {
		return nil
	}
	for _, n := range keyedList {
		m, ok := n.(map[string]interface{})
		if !ok {
			log.Errorf("wrong keyed list entry type: %T", n)
			return nil
		}
		keyMatching := true
		// must be exactly match
		for k, v := range elem.Key {
			attrVal, ok := m[k]
			if !ok {
				return nil
			}
			if v != fmt.Sprintf("%v", attrVal) {
				keyMatching = false
				break
			}
		}
		if keyMatching {
			return m
		}
	}
	if !createIfNotExist {
		return nil
	}

	// Create an entry in keyed list.
	m := make(map[string]interface{})
	for k, v := range elem.Key {
		m[k] = v
		if vAsNum, err := strconv.ParseFloat(v, 64); err == nil {
			m[k] = vAsNum
		}
	}
	node[elem.Name] = append(keyedList, m)
	return m
}

// Helper function, should be relocated elsewhere
// or is there something already within gnmi to do just this?
func createUpdateNotification(update *pb.Update) *gnmi.Notification {
	updates := make([]*pb.Update, 0)
	updates = append(updates, update)

	// create gnmi notification
	return &pb.Notification{
		Timestamp: time.Now().Unix(),
		Update:    updates,
	}
}

func createDeleteNotification(deletedPath *pb.Path) *pb.Notification {
	deletedPaths := make([]*pb.Path, 0)
	deletedPaths = append(deletedPaths, deletedPath)

	// create gnmi notification with deleted paths
	return &pb.Notification{
		Timestamp: time.Now().Unix(),
		Delete:    deletedPaths,
	}
}

// checkHandlerPaths runs over a gnmi.Notification and checks if the paths of
// its containing Updates or Deletes are the part of the area of responsibility
// of the given handler.
func checkHandlerPaths(h handler.PathHandler, diffs []*pb.Notification) ([]*handler.HandlerJob, error) {
	jobs := make([]*handler.HandlerJob, 0)

	for topic := range h.GetPaths() {
		for _, diff := range diffs {
			var updates []*gnmi.Update
			for _, update := range diff.GetUpdate() {
				schemaPath, err := ygot.PathToSchemaPath(update.GetPath())
				if err != nil {
					return []*handler.HandlerJob{}, err
				}
				if strings.Contains(schemaPath, topic) {
					updates = append(updates, update)
				}
			}

			var deletes []*gnmi.Path
			for _, del := range diff.GetDelete() {
				schemaPath, err := ygot.PathToSchemaPath(del)
				if err != nil {
					return []*handler.HandlerJob{}, err
				}
				if strings.Contains(schemaPath, topic) {
					deletes = append(deletes, del)
				}
			}

			if updates != nil || deletes != nil {
				jobs = append(jobs, &handler.HandlerJob{
					Updates: updates,
					Deletes: deletes,
				})
			}
		}
	}

	return jobs, nil
}
